/**
 * This file is part of the "Deniable Encryption" release in support of the
 * Blackhat Asia 2014 talk:
 *    A. Trachtenberg, "Say It Ain't So - An Implementation of Deniable Encryption".
 * 
 * The code in this file may code under patent and is released under the
 * following agreement:
 * http://nislab.bu.edu/wp-content/uploads/2014/03/Patented-Software-Release-Agreement.pdf
 */

/* 
 * File:   main.cpp
 * Author: Ari Trachtenberg
 * 
 * REQUIREMENTS:
 * *  Input files do not contain characters with ASCII values < Deniable::CHAR_THRESHOLD
 *
 * Created on November 14, 2013, 10:14 PM
 */

#include <cstdlib>
#include <cstdio>
#include <set>
#include "getopt.h"

#include "Auxiliary.h"
#include "Logger.h"
#include "Deniable.h"

using namespace std;

void usage(string cmd) {
    Logger::error_and_quit("Usage: " + cmd + "\n"
            + "   -i | --input   <filename> (may be repeated for encryption)" + "\n"
            + "   [-o | --output] <filename>" + "\n"
            + "   [-f | --keyfile] <filename>" + "\n"
            + "   [--encrypt | --decrypt]" + "\n"
            + "   [-? | --help]" + "\n"
            + "   *DECRYPTION*" + "\n"
            + "   [-k | --key]    <key string>" + "\n"
            + "   *ENCRYPTION*" + "\n"
            + "   [-e | --perr]   <error probability>" + "\n"
            + "   [-s | --sec]    <security paramater>" + "\n"
            + "---" + "\n"
            + "ENCRYPT mode:" + "\n"
            + "  Encrypt given input files into a ciphertext file (or stdout)," + "\n"
            + "    and one key per input file." + "\n"
            + "    * The input file must not contain null characters (\0)" + "\n"
            + "    * The keyfile is name of the file in which to store keys" + "\n"
            + "    * Optional arguments control the length of key needed for decryption." + "\n"
            + "    * Security parameter (epsilon in literature) trades off data security for key length" + "\n"
            + "DECRYPT mode:" + "\n"
            + "  Decrypt the given input file with the given key to the specified output file (or stdout)." + "\n"
    );
}

/** FOR TESTING
void gen_random(char *s, const int len) {
    static const char alphanum[] =
            "0123456789"
            "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
            "abcdefghijklmnopqrstuvwxyz";

    for (int i = 0; i < len; ++i) {
        s[i] = alphanum[rand() % (sizeof (alphanum) - 1)];
    }

    s[len] = 0;
}

string randStr(int len) {
    string ret = "";
    while (len-- > 0)
        ret += (char) rand() % 256;
    return ret;
}
 */

int main(int argc, char** argv) {
    // CONSTANTS
    const int DECRYPT = 0;
    const int ENCRYPT = 1;

    // VARIABLES
    string output; // the output
    string outputFile = ""; // possible encoded and decoded files
    string keyFile = "";    // filename where keys should be saved
    vector<string> inputFile; // possible decoy files
    string key = ""; // the decoding key
    double perr = 1E-10; // error probability
    int sec = -1; // security of encryption
    int option_index = 0;
    int mode_flag; // encode or decode

    // read in command-line options
    struct option long_options[] = {
        {"input", required_argument, 0, 'i'}, // input file
        {"output", required_argument, 0, 'o'}, // output file
        {"key", optional_argument, 0, 'k'}, // a key to use for decoding mode only
        {"keyfile", required_argument, 0, 'f'},  // filename for saving keys

        // mode
        {"decrypt", no_argument, &mode_flag, DECRYPT},
        {"encrypt", no_argument, &mode_flag, ENCRYPT},

        // sync parameters
        {"sec", required_argument, 0, 's'}, // security parameter
        {"perr", required_argument, 0, 'e'}, // maximum probability of error allowed
    };

    int choice = 0;
    do {
        switch (choice) {
            case 0: break; // long option without a short version ?
            case 'i': inputFile.push_back(optarg);
                break; // store an encoded filename
            case 'o': outputFile = optarg;
                break; // store a decoded filename
            case 'k': if (optarg != NULL)
                    key = optarg;
                else { // get the key from stdin
                    cout << "Please enter the decryption key: ";
                    cin >> key;
                }
                break; // the decryption key
            case 'f': keyFile = optarg; // the file keys should go
                break;
            case 's': sec = atoi(optarg);
                if (sec < 0)
                    Logger::error_and_quit("ERR:  Security parameter must be non-negative");
                break; // the security parameter
            case 'e': perr = strtod(optarg, NULL);
                break; // error probability
            default: //e.g. case '?':
                if (choice != '?')
                    cout << "Did not recognize choice: " << choice << endl;
                usage(argv[0]);
                exit(-1);
        }
        choice = getopt_long(argc, argv, "i:o:k:f:s:e:?",
                long_options, &option_index);
    } while (choice != -1);

    // Read in an input file
    if (inputFile.size() == 0) {
        cout << "No input file(s) specified!" << endl;
        usage(argv[0]);
    }

    if (mode_flag == ENCRYPT) {
        pair<string, vector<string> > result;
        try {
            result = Deniable::encrypt(inputFile, perr, sec);
        } catch (Deniable::EncryptionFailure def) {
            Logger::error_and_quit("Encryption failed due to \""+def.reason+"\"");
        }
        output = result.first;
        string keys="";
        for (int ii = 0; ii < result.second.size(); ii++) // print out the keys
            keys += "key #" + toStr(ii) + ": " + result.second[ii] + "\n";
        if (keyFile == "") // i.e. no output file
            cout << keys;  // write to stdout
        else
            Deniable::writeFile(keyFile, keys); // write to the specified file
        cout << "Files successfully encrypted" << endl;
    } else { // DECODE
        if (key == "") {
            cout << "No key specified for decryption!";
            usage(argv[0]);
        }
        try {
            output = Deniable::decrypt(key, inputFile[0]);
        } catch (Deniable::DecryptionFailure def) {
            cout << "Decryption failed with key " << def.key << " on file " << def.ciphertext << endl;
            exit(-1);
        }
        output = output.substr(1); // get rid of the leading \0 that marks the beginning of the string
    }
    // 4. Output the result
    if (outputFile == "") { // i.e. no output file
        cout << "Output was: " << output << endl;
        exit(0);
    } else { // write to a file
        Deniable::writeFile(outputFile, output);
    }
    return 0;
}

